/******************************************************************************
 * Copyright (c) 2013, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/
#include <xdc/std.h>
#include "ti/fftlib/src/common/fft_common.h"
#include "ti/fftlib/src/common/fft_edma.h"
#include "ti/fftlib/src/fft_sp_1d_r2c_batch/fft_sp_1d_r2c_batch.h"
#include <c6x.h>
extern cregister volatile unsigned int DNUM;

typedef struct {
  int              state;
  FFT_EDMA_Struct  edma;
} edmaGlobalState_t;

/* Hopefully cache will save us... 
 * It does not. Plan may only be created one at a time. 
 */
float *ptr_x_i;
float *ptr_y_i;
float *ptr_x_i_work;
float *ptr_y_i_work;
float *ptr_w_i;
float *ptr_bn_i;
float *ptr_w_i_work;
float *ptr_bn_i_work;
float *ptr_workbuf;

/* Every core should get this same base ptr, so there should be no issue */
char *g_fft_edma_state;

#define MAXN (8*1024)


#define FFTBUF_X_WORK_SIZE (8*MAXN)
#define FFTBUF_Y_WORK_SIZE (16*MAXN)
#define FFTBUF_W_WORK_SIZE (6*MAXN + 256)
#define FFTBUF_BN_SIZE (4*MAXN)
#define FFTBUF_WORKBUF_SIZE (8*MAXN)
#define FFTBUF_POOL_SIZE  (FFTBUF_X_WORK_SIZE + FFTBUF_Y_WORK_SIZE + FFTBUF_W_WORK_SIZE + FFTBUF_BN_SIZE + FFTBUF_WORKBUF_SIZE)


#define FFT_CACHE_ALIGNMENT  (0x80)
static inline int fft_cache_align(int size)
{
  return ((size + FFT_CACHE_ALIGNMENT - 1) & ~(FFT_CACHE_ALIGNMENT - 1));

}

void fft_memory_request (int nbufs, FFTmemBuffer_t *bufs)
{
  bufs[0].base = ptr_x_i;
  bufs[1].base = ptr_y_i;
  bufs[2].base = ptr_w_i;
  bufs[3].base = ptr_bn_i;
  bufs[4].base = ptr_x_i_work;
  bufs[5].base = ptr_y_i_work;
  bufs[6].base = ptr_w_i_work;
  bufs[7].base = ptr_bn_i_work;
  bufs[8].base = ptr_workbuf;
} 

void fft_memory_release (int nbufs, FFTmemBuffer_t *bufs) { }

void fft_assert(int statement, int node_id, const char *error)
{
    volatile int dbg_halt = 1;
    edmaGlobalState_t *state_ptr = (edmaGlobalState_t *)&g_fft_edma_state[DNUM * fft_cache_align(sizeof(edmaGlobalState_t))];


    if (!statement) {
      state_ptr->state = (unsigned int)error;
      while(dbg_halt);
    }
}

void *fft_assign_edma_resources(void)
{
  edmaGlobalState_t *state_ptr = (edmaGlobalState_t *)&g_fft_edma_state[DNUM * fft_cache_align(sizeof(edmaGlobalState_t))];
  /*
   * The edmaInstances are indexes into the C6678_config[] array defined in
   * fft_c6678_config, which is used to specify how EDMA resources are
   * divided between cores.
   */
  if ( state_ptr->state == 0 )
  {
    while ( state_ptr->edma.num_channels < FFT_NUM_EDMA_CH )
    {
      state_ptr->edma.channel[state_ptr->edma.num_channels] = EdmaMgr_alloc(FFT_MAX_EDMA_LINKS);

      fft_assert( (state_ptr->edma.channel[state_ptr->edma.num_channels] != NULL), 0, "EdmaMgr_alloc() failed ");

      state_ptr->edma.num_channels++;
    }
    state_ptr->state = 1;
  }
  return((void *)&state_ptr->edma);
}

void fft_free_edma_resources(FFT_EDMA_Struct *edmaState)
{
  edmaGlobalState_t *state_ptr = (edmaGlobalState_t *)&g_fft_edma_state[DNUM * fft_cache_align(sizeof(edmaGlobalState_t))];
  int ret_val;
/*
 * The edmaInstances are indexes into the C6678_config[] array defined in
 * fft_c6678_config, which is used to specify how EDMA resources are
 * divided between cores.
 */
  fft_assert((edmaState == &state_ptr->edma), 0, "Obtained incorrect FFT_EDMA_Struct");

  if (state_ptr->state == 1)
  {
    while ( state_ptr->edma.num_channels > 0 )
    {
      ret_val = EdmaMgr_free(state_ptr->edma.channel[state_ptr->edma.num_channels - 1]);
      fft_assert( (ret_val == EdmaMgr_SUCCESS), 0, "EdmaMgr_free failed ");
      state_ptr->edma.num_channels--;
    }
    state_ptr->state = 0;
  }
}

void fft_set_work_bufs(fft_plan_t *plan, float *buf_pool)
{
    plan->u.sp_1d_r2c_batch_e.in_w  = &buf_pool[0];
    plan->u.sp_1d_r2c_batch_e.out_w = &plan->u.sp_1d_r2c_batch_e.in_w[FFTBUF_X_WORK_SIZE/4];
    plan->u.sp_1d_r2c_batch_e.tw_w  = &plan->u.sp_1d_r2c_batch_e.out_w[FFTBUF_Y_WORK_SIZE/4];
    plan->u.sp_1d_r2c_batch_e.Bn    = &plan->u.sp_1d_r2c_batch_e.tw_w[FFTBUF_W_WORK_SIZE/4];
    plan->u.sp_1d_r2c_batch_e.Bn_w  = &plan->u.sp_1d_r2c_batch_e.Bn[FFTBUF_BN_SIZE/4];
}

void fft_acquire_edma_resources(int core_id, char *global_state)
{
  edmaGlobalState_t *state_ptr = (edmaGlobalState_t *)&global_state[core_id * fft_cache_align(sizeof(edmaGlobalState_t))];
  int ret_val;

  if (state_ptr->state == 0)
  {
    while ( state_ptr->edma.num_channels < FFT_NUM_EDMA_CH )
    {
      state_ptr->edma.channel[state_ptr->edma.num_channels] = EdmaMgr_alloc(FFT_MAX_EDMA_LINKS);

      fft_assert( (state_ptr->edma.channel[state_ptr->edma.num_channels] != NULL), 0, "EdmaMgr_alloc() failed ");

      state_ptr->edma.num_channels++;
    }
    state_ptr->state = 1;
  }
}

void fft_release_edma_resources(int core_id, char *global_state)
{
    edmaGlobalState_t *state_ptr = (edmaGlobalState_t *)&global_state[core_id * fft_cache_align(sizeof(edmaGlobalState_t))];
    int ret_val;

    if (state_ptr->state == 1)
    {
      while ( state_ptr->edma.num_channels > 0 )
      {
        ret_val = EdmaMgr_free(state_ptr->edma.channel[state_ptr->edma.num_channels - 1]);
        fft_assert( (ret_val == EdmaMgr_SUCCESS), 0, "EdmaMgr_free failed ");
        state_ptr->edma.num_channels--;
      }
    }
    
    state_ptr->state = 0;
}


void FFT_plan_facade(int N, int BATCH, int type, int mode, char *plan,
       float *x_i, float *y_i, float *w_i, float *fftbufs_pool, char *globalState)
{
    fft_callout_t               plan_fxns;
    fft_plan_t p;

    g_fft_edma_state = globalState;

    ptr_x_i      = x_i;
    ptr_y_i      = y_i;
    ptr_w_i      = w_i;

    plan_fxns.memoryRequest   = fft_memory_request;
    plan_fxns.ecpyRequest     = fft_assign_edma_resources;
    plan_fxns.memoryRelease = fft_memory_release;
    plan_fxns.ecpyRelease   = NULL; //fft_free_edma_resources;

    fft_acquire_edma_resources(DNUM,globalState);

#if 1
    p = fft_sp_plan_1d_r2c_batch(N, BATCH,mode, plan_fxns);
#else
    p = fft_sp_plan_1d_r2c_batch_dsplib(N, BATCH,mode, plan_fxns);
#endif
    memcpy(plan, (char*)&p, sizeof(fft_plan_t));
    fft_set_work_bufs((fft_plan_t*)plan,fftbufs_pool);
}

void FFT_execute_facade(char *plan, float *fftbufs_pool, char *globalState)
{
    UInt hwi_state;

    g_fft_edma_state = globalState;

    fft_set_work_bufs((fft_plan_t*)plan, fftbufs_pool);

    fft_execute(*(fft_plan_t*)plan);

}

void FFT_destroy_plan_facade(char *plan, char *globalState)
{
    g_fft_edma_state = globalState;

    fft_destroy_plan(*(fft_plan_t*)plan);
}


